@@ -1,5 +1,6 @@
 # pg.py
 # Written by D'Arcy J.M. Cain
+# $Id: pg.py,v 1.28 2005/01/08 15:04:35 darcy Exp $
 
 # This library implements some basic database management stuff.  It
 # includes the pg module and builds on it.  This is known as the
@@ -40,7 +41,7 @@ def _quote(d, t):
 	if t in ('inet', 'cidr') and d == '': return "NULL"
 
 	return "'%s'" % string.strip(re.sub("'", "''", \
-							 re.sub("\\\\", "\\\\\\\\", "%s" % d)))
+							re.sub("\\\\", "\\\\\\\\", "%s" % d)))
 
 class DB:
 	"""This class wraps the pg connection type"""
@@ -140,25 +141,37 @@ class DB:
 			self.__attnames = newattnames
 			return
 		elif newattnames:
-			raise error, "If supplied, newattnames must be a dictionary"
+			raise ProgrammingError, \
+					"If supplied, newattnames must be a dictionary"
 
 		# May as well cache them
 		if self.__attnames.has_key(cl):
 			return self.__attnames[cl]
 
+		# check for schema name
+		if cl.find('.') == -1:
+			schema = 'public'
+			table = cl
+		else:
+			try: schema, table = cl.split('.')
+			except ValueError, err:
+				raise ProgrammingError('Invalid class %s' % cl)
+
 		query = """SELECT pg_attribute.attname, pg_type.typname
-					FROM pg_class, pg_attribute, pg_type
+					FROM pg_class, pg_attribute, pg_type, pg_namespace
 					WHERE pg_class.relname = '%s' AND
+						pg_namespace.nspname = '%s' AND
 						pg_attribute.attnum > 0 AND
 						pg_attribute.attrelid = pg_class.oid AND
 						pg_attribute.atttypid = pg_type.oid AND
-						pg_attribute.attisdropped = 'f'"""
+						pg_class.relnamespace = pg_namespace.oid AND
+						pg_attribute.attisdropped = 'f'""" % (table, schema)
 
 		l = {}
-		for attname, typname in self.db.query(query % cl).getresult():
+		for attname, typname in self.db.query(query).getresult():
 			if re.match("^interval", typname):
-				l[attname] = 'text'
-			if re.match("^int", typname):
+				l[attname] = 'date'
+			elif re.match("^int", typname):
 				l[attname] = 'int'
 			elif re.match("^oid", typname):
 				l[attname] = 'int'
@@ -222,7 +235,7 @@ class DB:
 		self._do_debug(q)
 		res = self.db.query(q).dictresult()
 		if res == []:
-			raise error, \
+			raise DatabaseError, \
 				"No such record in %s where %s is %s" % \
 								(cl, keyname, _quote(k, fnames[keyname]))
 			return None
@@ -243,13 +256,10 @@ class DB:
 				l.append(_quote(a[f], fnames[f]))
 				n.append(f)
 
-		try:
-			q = "INSERT INTO %s (%s) VALUES (%s)" % \
-				(cl, string.join(n, ','), string.join(l, ','))
-			self._do_debug(q)
-			a['oid_%s' % cl] = self.db.query(q)
-		except:
-			raise error, "Error inserting into %s: %s" % (cl, sys.exc_value)
+		q = "INSERT INTO %s (%s) VALUES (%s)" % \
+			(cl, string.join(n, ','), string.join(l, ','))
+		self._do_debug(q)
+		a['oid_%s' % cl] = self.db.query(q)
 
 		# reload the dictionary to catch things modified by engine
 		# note that get() changes 'oid' below to oid_table
@@ -260,15 +270,15 @@ class DB:
 	# Update always works on the oid which get returns if available
 	# otherwise use the primary key.  Fail if neither.
 	def update(self, cl, a):
-		self.pkey(cl)		# make sure we have a self.__pkeys dictionary
-
 		foid = 'oid_%s' % cl
 		if a.has_key(foid):
 			where = "oid = %s" % a[foid]
-		elif self.__pkeys.has_key(cl) and a.has_key(self.__pkeys[cl]):
-			where = "%s = '%s'" % (self.__pkeys[cl], a[self.__pkeys[cl]])
 		else:
-			raise error, "Update needs primary key or oid as %s" % foid
+			try: pk = self.pkey(cl)
+			except: raise ProgrammingError, \
+					"Update needs primary key or oid as %s" % foid
+
+			where = "%s = '%s'" % (pk, a[pk])
 
 		v = []
 		k = 0
@@ -281,13 +291,9 @@ class DB:
 		if v == []:
 			return None
 
-		try:
-			q = "UPDATE %s SET %s WHERE %s" % \
-							(cl, string.join(v, ','), where)
-			self._do_debug(q)
-			self.db.query(q)
-		except:
-			raise error, "Can't update %s: %s" % (cl, sys.exc_value)
+		q = "UPDATE %s SET %s WHERE %s" % (cl, string.join(v, ','), where)
+		self._do_debug(q)
+		self.db.query(q)
 
 		# reload the dictionary to catch things modified by engine
 		if a.has_key(foid):
@@ -311,12 +317,7 @@ class DB:
 	# one day we will be testing that the record to be deleted
 	# isn't referenced somewhere (or else PostgreSQL will)
 	def delete(self, cl, a):
-		try:
-			q = "DELETE FROM %s WHERE oid = %s" % (cl, a['oid_%s' % cl])
-			self._do_debug(q)
-			self.db.query(q)
-		except:
-			raise error, "Can't delete %s: %s" % (cl, sys.exc_value)
-
-		return None
+		q = "DELETE FROM %s WHERE oid = %s" % (cl, a['oid_%s' % cl])
+		self._do_debug(q)
+		self.db.query(q)
 
